;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; STEGA v2.01 - Tiny 4058-bytes Steganography Program
; Copyright(c) 2006 by Asko Vuori, Finland
;
; Released into the FreeWare - FREE TO USE.
; Please distribute this stuff as you got it.
; Don't distribute modified files. Thanx.
;
; Program to hide file inside BMP/GIF/PCX/TGA picture or VOC/WAV sound file
; or RAW 8-bit data file or LST/TXT text file.
; It's recomended that the picture is GrayScale or TrueColor.
; Written in 8086 assembly.
;
; Usage:
;   STEGA +|-[ABs,cHtLwPs,b,dRSUs,e] Host.Gif Hide.Ext
;
; Compile with FASM
;
; Version history:
;   v1.01 - (SBMP/SGIF/SPCX/STGA/STXT) by Asko Vuori, Finland (20.Dec.2000)
;   v2.00 - 4058 bytes by Asko Vuori, Finland (25.Nov.2001)
;   v2.01 - converted to FASM syntax (29.Oct.2006)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;       title   STEGA - Tiny Steganography Program

include 'macro\struct.inc'		;Use fasmw struct macro

; Constants
BUFFERSIZE	equ 2048		;Idea/Input/Output buffers
PassLen 	equ 128 		;Passphrase length

None		equ 0			;Error numbers (maybe future)

; Gif Constants
NOCODE		equ 0FFFFh
MAXBITS 	equ 12
HASHSIZE	equ 1 shl MAXBITS	;2^12=4096

; BMP header structure (54 bytes)
struct BmpHeaderType
  BmpSign	    dw	?
  BmpFileSize	    dd	?
  BmpRes1	    dw	?
  BmpRes2	    dw	?
  BmpStartPic	    dd	?
  BmpInfoSize	    dd	?
  BmpWidth	    dd	?
  BmpDepth	    dd	?
  BmpColorPlanes    dw	?
  BmpBitsPerPixel   dw	?
  BmpCompresson     dd	?
  BmpSizeImage	    dd	?
  BmpXPelsPerMeter  dd	?
  BmpYPelsPerMeter  dd	?
  BmpColorUsed	    dd	?
  BmpColorsImport   dd	?
ends

struct GifHeaderType
  ghSig 	    db	6 dup(?)
  ghScreenWidth     dw	?
  ghScreenDepth     dw	?
  ghFlags	    db	?
  ghBackGround	    db	?
  ghAspect	    db	?
ends

struct GifImageBlock
  ibLeft	    dw	?
  ibTop 	    dw	?
  ibWidth	    dw	?
  ibDepth	    dw	?
  ibFlags	    db	?
ends

struct GifControlBlock
  cbBlockSize	    db	?
  cbFlags	    db	?
  cbDelay	    dw	?
  cbTransparentCol  db	?
  cbTerminator	    db	?
ends

struct GifPlainText
  ptBlockSize	    db	?
  ptLeft	    dw	?
  ptTop 	    dw	?
  ptGridWidth	    dw	?
  ptGridHeight	    dw	?
  ptCellWidth	    db	?
  ptCellHeight	    db	?
  ptForeColor	    db	?
  ptBackColor	    db	?
ends

struct GifApplication
  gaBlockSize	    db	?
  gaApplString	    db	8 dup(?)
  gaAuthentication  db	3 dup(?)
ends

; PCX header structure (128 bytes)
struct PcxHeaderType
  PcxManufacture    db	?
  PcxVersion	    db	?
  PcxEncoding	    db	?
  PcxBitsPerPixel   db	?
  PcxXMin	    dw	?
  PcxYMin	    dw	?
  PcxXMax	    dw	?
  PcxYMax	    dw	?
  PcxWidth	    dw	?
  PcxDepth	    dw	?
  PcxPalette	    db	49 dup(?)
  PcxColorPlanes    db	?
  PcxBytesPerLine   dw	?
  PcxPaletteInfo    dw	?
  PcxFiller	    db	58 dup(?)
ends

; TARGA header structure (18 bytes)
struct TgaHeaderType
  TgaTextSize	    db	?	;Number of bytes in Ident Field
  TgaMapType	    db	?	;Color map type (0 for none)
  TgaDataType	    db	?	;Image type (2 is unmapped RGB)
  TgaMapOrig	    dw	?	;Color map origin
  TgaMapLength	    dw	?	;Color map length in # of colors
  TgaCMapBits	    db	?	;Number of bits per color map entry
  TgaXOffset	    dw	?
  TgaYOffset	    dw	?
  TgaX		    dw	?
  TgaY		    dw	?
  TgaBitsPerPixel   db	?	;Number of bits per pixel in picture
  TgaImType	    db	?
ends

; VOC header structure (26 bytes)
struct VocHeaderType
  VocID 	    db	19 dup(?);"Creative Voice File"
  VocEofChr	    db	?	;1Ah (Eof to abort printing of file)
  VocOfsData	    dw	?	;Offset of first datablock in .VOC file
  VocVersion	    dw	?	;Version number (minor,major)
  VocCompVers	    dw	?	;2's Comp of Ver. # + 1234h
ends

; RIFF header structure (12 bytes)
struct RiffHeaderType
  RiffID	    db	4 dup(?);"RIFF"
  RiffLen	    dd	?	;The length of the data in the next chunk
  RiffDataID	    db	4 dup(?);"WAVE" (future "AVI "?)
ends

; WAV header structure (44 bytes)
struct WavHeaderType
  WavFileID	    RiffHeaderType
  WavFormID	    db	4 dup(?);"fmt "
  WavFormLen	    dd	?	;Length of data in the format chunk
  WavFormTag	    dw	?
  WavChannels	    dw	?	;Number of channels, 1=mono, 2=stereo
  WavSamplesPerSec  dd	?	;Playback frequency
  WavAvgBytesPerSec dd	?
  WavBlockAlign     dw	?
  WavFormatSpecific dw	?	;Format specific data area
  WavDataID	    db	4 dup(?);"data"
  WavDataLen	    dd	?	;Length of data in the dData field
ends

struct ExtType
  FileExt	    db	4 dup(?)
  ExtOffset	    dw	?
ends

	org	0080h
CmdLine = $
	org	0100h

StegaMain:
	cld
	xor	cx,cx
	mov	bx,CmdLine
SkipSpc:inc	bx
	mov	al,[bx]
	cmp	al,0Dh
	jz	Str0
	cmp	al,20h
	jbe	SkipSpc
	cmp	al,'+'
	jz	SetDM
	cmp	al,'-'
	jnz	NotDM
SetDM:	mov	[DecodeMode],al
	mov	ch,02h
	inc	bx
TestNu: mov	al,[bx]
	and	al,0DFh 		;Convert to uppercase
	cmp	al,'A'
	jb	Skip2
	call	TestOptions
	jmp	TestNu
NotDM:	cmp	cl,1
	ja	SkipChr
	jz	Param2
	mov	[HostName],bx
Param2: mov	[HideName],bx
	inc	cx
SkipChr:inc	bx
Skip2:	mov	al,[bx]
	cmp	al,20h
	ja	SkipChr
Str0:	mov	[bx],bh
	cmp	al,0Dh			;Is end of CmdLine?
	jnz	SkipSpc
	cmp	cl,1
	jnz	HNameOk
	call	IsHide
	jnz	HNameOk
	mov	[DecodeMode],'#'	;Set clear mode flag
	mov	[WhatLength],0
	jmp	NoKey
HNameOk:test	cl,ch			;Is enough parameters?
	jz	EndTxt
	cmp	[IsIdea],0
	jz	NoKey
	call	GetIDEAKey
NoKey:	call	TestExtension
	jc	EndErr			;Is file extension ok?
	call	IsHide			;If hide '+' test HideSize
	jz	IsHiSi0
	cmp	[WhatLength],0		;If 0 then no file length
	jz	TheEnd
IsHiSi0:call	IsHideSize0		;If HideSize=0 all done
	jz	TheEnd
	mov	dx,TxtTooBig
EndErr: mov	ah,09h			;Dos WriteString
	int	21h
EndTxt: mov	dx,TxtInfo
	mov	ah,09h			;Dos WriteString
	int	21h
TheEnd: mov	cx,GifStack - Password
	mov	di,Password+1
	rep	stosb			;Clear Passphrase and buffers
	int	20h			;Exit to Dos
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
TestOptions:
	cmp	al,'A'			;Is append file?
	jnz	NotApp			;Future '+As = HideFile StartPos'
	mov	[WriteMode],al
NotApp: inc	bx			;Update CmdLine position
	cmp	al,'B'			;Is bit position?
	jnz	NotBit
	call	GetNumber
	jnz	GeBit1
	dec	ax
	cmp	al,8
	jnb	GeBit1
	inc	ax
	mov	[Roll],al
GeBit1: call	GetNumberAfterComma
	jnz	ClrRollCount
	dec	ax
	cmp	al,8
	jnb	ClrRollCount
	mov	ah,8
	sub	ah,[Roll]
	cmp	al,ah
	jbe	GeBit3
ClrRollCount:
	mov	al,0
GeBit3: mov	[RollCount],al
	ret
NotBit: cmp	al,'H'			;Is header type?
	jnz	NotHead
	call	GetNumber
	jnz	TeOpRe
	cmp	al,2
	ja	TeOpRe
	add	al,al
	mov	[WhatLength],al 	;0=no length, 2=16-bit, 4=32-bit
	ret
NotHead:cmp	al,'L'			;Is line length?
	jnz	NotLen
	call	GetNumber
	jnz	TeOpRe
	mov	[LineLen],ax
	ret
NotLen: cmp	al,'P'			;Get position?
	jnz	NotPos
	call	GetNumber
	jnz	GePos1
	mov	[DummyCounter],ax
GePos1: call	GetNumberAfterComma
	jnz	GePos2
	test	ax,ax
	jz	GePos2
	mov	[DefCount],ax
GePos2: call	GetNumberAfterComma
	jz	GePos3
	mov	ax,1
GePos3: mov	[DefDummyCount],ax
GePosRe:ret
NotPos: cmp	al,'R'			;Is reverse?
	jnz	NotRev
	xor	byte [RclPos+1],08h	;Selfmodify code to swap rcl/rcr
NotRev: cmp	al,'S'			;Is scrambled?
	jnz	NotIdea
	mov	[IsIdea],al
NotIdea:cmp	al,'U'			;Get use range?
	jnz	TeOpRe
	call	GetNumber
	jnz	ClrUse
	mov	[UseMin],al
	call	GetNumberAfterSwap
	mov	[UseMax],al
	jz	TeOpRe
ClrUse: mov	[UseFlag],'x'
TeOpRe: ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Param:   bx = UseMin and UseMax
;         [ds:si] = Byte to test
;Return:  CF is set if inside range
IsInRange:
	cmp	bl,bh			;Is exclude range?
	jb	IsInRa2
	cmp	[si],bl
	jb	IsInRa3
IsInRa1:stc
	ret
IsInRa2:cmp	[si],bl
	jb	IsInRa4
IsInRa3:cmp	[si],bh
	jbe	IsInRa1
IsInRa4:clc
	ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
GetNumberAfterSwap:
	mov	al,[bx]
	mov	[UseFlag],al
	cmp	al,'='			;Is swap mode?
	jz	GeNuAf1
GetNumberAfterComma:
	cmp	byte [bx],','
	jnz	GeNuRet
GeNuAf1:inc	bx
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Param:   [ds:bx] -> Ascii number
;Return:  ax = Number (ZF is set if number OK)
;Destroy: dx,si
GetNumber:
	push	cx
	mov	cx,000Ah
	mov	dh,0FFh 		;Error flag
	xor	si,si
GeNu1:	mov	al,[bx] 		;First digit of number
	cmp	al,'$'
	jnz	GeNu2
	mov	cl,10h
	inc	bx
	jmp	GeNu1
GeNu2:	cmp	al,'9'
	jbe	GeNu4
	cmp	cl,10h
	jnz	GeNu5
	and	al,0DFh 		;Convert to uppercase
	cmp	al,'A'
	jb	GeNu5
	sub	al,7
GeNu4:	sub	al,'0'
	test	al,0F0h
	jnz	GeNu5
	inc	bx
	cbw
	xchg	ax,si
	mul	cx
	add	si,ax
	jnc	GeNu1
GeNu5:	xchg	ax,si
	test	dh,dh			;ZF is set if number OK
	pop	cx
GeNuRet:ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
TestExtension:
	xor	di,di
	mov	si,[HostName]
TeEx0:	mov	[PureHostName],si	;Host file name without path
TeEx1:	lodsb
	cmp	al,'\'
	jz	TeEx0
	cmp	al,'.'
	jnz	TeEx2
	mov	di,si
TeEx2:	cmp	al,0
	jnz	TeEx1
	test	di,di
	jz	TeEx6
	mov	si,ExtList
TeEx3:	xor	bx,bx
TeEx4:	mov	al,[di+bx]
	and	al,0DFh 		;Convert to uppercase
	cmp	al,[si+bx+ExtType.FileExt]
	jnz	TeEx5
	inc	bx
	cmp	byte [si+bx+ExtType.FileExt],0
	jnz	TeEx4
	push	[si+ExtType.ExtOffset]
	call	OpenFiles
	pop	ax
	mov	dx,TxtOpenError
	jc	TeExErr
	call	ax
	rcl	[FileErr],1
	call	CloseFiles
	mov	dx,TxtFileError
	cmp	[FileErr],None
	jnz	TeExErr
	clc
	ret
TeEx5:	add	si,sizeof.ExtType
	cmp	si,ExtListEnd
	jb	TeEx3
TeEx6:	mov	dx,TxtUnknownForm
TeExErr:stc
	ret
ExtList:ExtType 'BMP',DoBmpFile
	ExtType 'GIF',DoGifFile
	ExtType 'LST',DoLstFile
	ExtType 'PCX',DoPcxFile
	ExtType 'RAW',DoRawFile
	ExtType 'TGA',DoTgaFile
	ExtType 'TXT',DoTxtFile
	ExtType 'VOC',DoVocFile
	ExtType 'WAV',DoWavFile
ExtListEnd:
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Used registers:
;  si=Input byte position
;  cx=BmpLine / LinePos
;  bp=BytesPerLine
DoBmpFile:
	mov	cx,sizeof.BmpHeaderType
	call	ReadHeader
	jc	BmpErr
	cmp	[BmpHeader.BmpSign],'BM' ;Check if Id = 'BM'
	jnz	BmpErr
	cmp	word [BmpHeader.BmpCompresson],0
	jnz	BmpErr
	call	WriteHeader
	jnc	DoBmp1
BmpErr: stc
	ret
DoBmp1: mov	ax,[BmpHeader.BmpBitsPerPixel]
	mul	word [BmpHeader.BmpWidth]
	mov	cx,8
	cmp	cx,dx
	jbe	BmpErr
	div	cx
	xchg	bp,ax
	mov	ax,word [BmpHeader.BmpStartPic]
	mov	dx,word [BmpHeader.BmpStartPic+2]
	sub	ax,sizeof.BmpHeaderType
	sbb	dx,0
	call	CopyBlockBig		;Seek over color palette
	jc	BmpErr
	mov	cx,word [BmpHeader.BmpDepth]
DoBmp2: push	cx
	mov	di,bp
	xor	dx,dx
	call	HandleBlock
	jc	DoBmp4
	xor	ax,ax
	sub	ax,bp
	and	ax,3
	call	CopyBlock
DoBmp4: pop	cx			;PicDepth
	jc	BmpErr
	call	IsAllDone
	loopne	DoBmp2			;Loop back until all done
	clc
	ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
DoGifFile:
	mov	[gCon.cbFlags],0	;Clear transparent flag
	mov	cx,sizeof.GifHeaderType
	call	ReadHeader
	jc	DoGifEr
	cmp	word [GifHeader.ghSig],'GI' ;Make sure it's a Gif file
	jnz	DoGifEr
	call	WriteHeader
	jc	DoGifEr
DoGif1: mov	al,[GifHeader.ghFlags]	;Get PaletteFlag
	call	GetVgaPalette
	jc	DoGifEr
DoGif2: call	CopyByte
	jbe	DoGifEr
	cmp	al,';'			;Is EndOfFile?
	jz	DoGifRe 		;Exit with clear CF
	cmp	al,'!'			;It's an extension block
	jnz	DoGif3
	call	SkipExtension
	jmp	IsGifEr
DoGif3: cmp	al,','			;It's an image block
	jnz	DoGif4
	call	UnpackImage
IsGifEr:jc	DoGifEr
DoGif4: cmp	[FileErr],None
	jz	DoGif2
DoGifEr:stc
DoGifRe:ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
UnpackImage:
	mov	cx,sizeof.GifImageBlock
	mov	dx,gifi
	call	GetBlock
	jc	UnImRet
	call	PutBlock
	jc	UnImRet
	;Get the local color map if there is one
	mov	al,[gifi.ibFlags]	;Get PaletteFlag
	call	GetVgaPalette
	jc	UnImRet
	call	CopyByte		;First we get the intial code size
	jbe	UnImRet
	mov	[InitCodeSize],al
	xor	ax,ax
	mov	[IBitsLeft],ax		;All bits are zero
	mov	[IBitCount],al		;so IBitCount is zero
	mov	[GifBufCount],ax	;Read buffer is empty
	mov	[GifBufPos],ax
	mov	[OBitOffset],ax 	;Clear OBitOffset of WrBuffer
	call	IsSeek
	jz	UnpackGif		;Direct unpacking and return
	mov	[OIsFirst],1
	call	InitTable		;Initialize the table and some vars
	mov	ax,[OClearCode] 	;Write a clear code
	call	WriteCode
	call	UnpackGif
	jc	UnImRet
	mov	ax,[OPrefix]		;Write last code
	call	WriteCode
	mov	ax,[OEofCode]		;Write Eof code
	call	WriteCode
	mov	ax,[OBitOffset]
	add	ax,7			;If extra bits, make sure they get
	mov	cl,3			;written
	shr	ax,cl			;Convert bits to bytes
	clc
	jz	UnImRet
Flush:	call	PutByte 		;Write the buffer length
	jc	UnImRet
	xchg	ax,cx
	mov	dx,WrBuffer
	call	PutBlock
UnImRet:ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Used registers:
;  bp=ReadMask
;  si='stack count'
UnpackGif:
DoClear:mov	cl,[InitCodeSize]	;Compute decompressor constant values,
	mov	ax,1			;based on the init code size
	mov	bp,ax
	shl	ax,cl
	mov	[IClearCode],ax
	inc	ax
	mov	[IEofCode],ax
	inc	ax
	mov	[IFreeCode],ax		;First free code
	inc	cx
	mov	[ICodeSize],cl
	shl	bp,cl
	dec	bp			;bp=ReadMask for current code size
	mov	[IOldCode],NOCODE
UGif1:	call	ReadCode		;Get the next code from the file
	jbe	UGifRet 		;Is error or Eof found?
	cmp	ax,[IClearCode]
	jz	DoClear
	mov	bx,ax			;If not a clear code then must be data
	xchg	ax,[IOldCode]		;Swap IOldCode and new value
	mov	si,GifStack		;Decompressor output 'stack count'
	;If greater or equal to IFreeCode, not in the hash table yet
	;repeat the last character decoded.
	mov	di,[IFreeCode]
	cmp	bx,di
	ja	UGifEr			;Something is missing
	jnz	UGif5
	cmp	ax,NOCODE
	jz	UGifRet 		;Exit with clear CF
	mov	bx,ax			;bx=IOldCode
	mov	dl,[IOldToken]
UGif4:	mov	[si],dl 		;Save byte in 'stack buffer'
	inc	si
	;Unless this code is raw data, pursue the chain pointed to by
	;bx through the hash table to its end; each code in the chain
	;puts its associated output code on the output queue.
UGif5:	cmp	bx,[IClearCode]
	jb	UGif6
	mov	dl,[bx+Suffix]
	add	bx,bx
	mov	bx,[bx+Prefix]
	jmp	UGif4
	;The last code in the chain is treated as raw data.
UGif6:	mov	[IOldToken],bl
	mov	[si],bl 		;Save byte in 'stack buffer'
	inc	si
	cmp	ax,NOCODE		;ax=IOldCode
	jz	UGif8
	;If we exceed the current bp=ReadMask value, increment the code size
	;unless it's already 12.
	;If it is, do nothing: the next code decompressed better be CLEAR
	cmp	di,bp
	jb	UGif7
	cmp	di,HASHSIZE-1		;If di=4095 then ICodeSize=12
	jnb	UGif8
	rcl	bp,1			;Update ReadMask
	inc	[ICodeSize]
UGif7:	mov	[di+Suffix],bl		;Build the hash table on-the-fly.
	add	di,di
	mov	[di+Prefix],ax
	inc	[IFreeCode]		;Point to the next slot in the table.
UGif8:	call	UnpackLiFo
	cmp	[FileErr],None
	jz	UGif1
UGifEr: stc
UGifRet:ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
UnpackLiFo:
	;Put the data out. It's been stacked LIFO, so deal with it that way.
UnLi1:	dec	si			;Update byte position
	test	[gCon.cbFlags],01h	;Is transparent flag?
	jz	UnLi2
	mov	al,[gCon.cbTransparentCol]
	cmp	al,[si]
	jz	UnLi3
UnLi2:	call	HandleByte
UnLi3:	call	IsSeek
	jz	UnLi4
	mov	al,[si]
	call	CompressByte
UnLi4:	cmp	si,GifStack
	ja	UnLi1
	ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
SkipExtension:
	call	CopyByte
	jbe	SkExRet
	mov	cx,sizeof.GifPlainText
	cmp	al,01h			;01h = Plain text descriptor
	jz	SkEx1
SNoTxt: cmp	al,0F9h 		;0F9h = Graphic control block
	jnz	SNoCon
	mov	cx,sizeof.GifControlBlock
	mov	dx,gCon
	call	GetBlock
	jc	SkExRet
	call	PutBlock
	ret
SNoCon: cmp	al,0FFh 		;0FFh = Application extension
	jnz	SkEx2
	mov	cx,sizeof.GifApplication
SkEx1:	xchg	ax,cx
	call	CopyBlock
	jc	SkExRet
SkEx2:	call	CopyByte		;0FEh = Comment extension
	jc	SkExRet 		;or <= 0FDh = Something else
	xor	ah,ah
	push	ax
	call	CopyBlock
	pop	ax
	jc	SkExRet
	test	ax,ax
	jnz	SkEx2
	clc
SkExRet:ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Return:  ax = ReadCode (ZF is set if Eof found)
;         CF is set on error
ReadCode:
	mov	ax,[IBitsLeft]
	mov	cx,[ICodeBitsW]
	shr	[IBitsLeft],cl
ReCo1:	cmp	ch,cl
	jnb	ReCo3
	mov	bx,[GifBufPos]
	cmp	bx,[GifBufCount]
	jb	ReCo2
	push	ax
	push	cx
	call	GetByte
	jb	ReCoEr
	xor	ah,ah
	xchg	ax,cx
	mov	dx,ReBuffer
	call	GetBlock
	jb	ReCoEr			;Can't read bytes, so exit (CF=Error)
	mov	[GifBufCount],ax
	pop	cx
	pop	ax
	jz	ReCoEx
	xor	bx,bx
ReCo2:	xor	dx,dx
	mov	dl,[bx+ReBuffer]
	inc	bx
	mov	[GifBufPos],bx
	mov	bx,dx
	push	cx
	sub	cl,ch
	shr	dx,cl
	mov	cl,ch
	shl	bx,cl
	or	ax,bx
	pop	cx
	mov	[IBitsLeft],dx
	add	ch,8
	jmp	ReCo1
ReCo3:	sub	ch,cl
	mov	[IBitCount],ch
	and	ax,bp			;Mask with ReadMask
	cmp	ax,[IEofCode]
ReCoEx: clc
	ret
ReCoEr: pop	cx
	pop	ax
	stc
	ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
InitTable:
	mov	ax,NOCODE		;Unused flag
	mov	di,hFirst
	mov	cx,(HASHSIZE * 5) / 2
	rep	stosw			;Clear tables
	mov	cl,[InitCodeSize]
	mov	ax,2
	shl	ax,cl
	mov	[OMaxCode],ax		;Set max code for current code size
	shr	ax,1
	mov	[OClearCode],ax
	inc	ax
	mov	[OEofCode],ax
	inc	ax
	mov	[OFreeCode],ax		;Set next free code to use
	inc	cx
	mov	[OCodeSize],cx		;Set first code size
	ret				;Done
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Param:   al=Byte to pack
CompressByte:
	shr	[OIsFirst],1		;Test and Clear flag
	jc	CoBy7			;Only CF for first byte
	mov	[OLastChar],al		;Save char in OLastChar
	mov	bx,[OPrefix]		;Get prefix code
	mov	cx,NOCODE
	add	bx,bx			;See if this pair in table
	mov	di,[bx+hFirst]		;Get first entry
	cmp	di,cx			;Has this code been used?
	jne	CoBy1			;Not equal means yes
	;First use of this prefix ? Not found, so add pair to table
	mov	di,[OFreeCode]		;Get code to use
	mov	[bx+hFirst],di		;Point first use to new entry
	jmp	CoBy3
CoBy1:	cmp	[di+hChar],al		;Is char the same?
	jne	CoBy2			;No.
	mov	ax,di			;Yes. New code in ax
	jmp	CoBy8
CoBy2:	mov	bx,di
	mov	di,[bx+di+hNext]	;Get next code
	cmp	di,cx			;More left with this prefix?
	jne	CoBy1			;Not equal means try again
	;Next use of this prefix ? Not found, so add pair to table
	add	bx,bx
	mov	di,[OFreeCode]		;Get code to use
	mov	[bx+hNext],di		;Point last use to new entry
CoBy3:	cmp	di,HASHSIZE		;Have we reached code limit?
	je	CoBy4			;Equal means yes, just return
	mov	[di+hChar],al		;Save suffix char
	mov	bx,di
	mov	[bx+di+hFirst],cx	;Initialize pointers
	mov	[bx+di+hNext],cx
	inc	[OFreeCode]		;Adjust next code
CoBy4:	push	di			;Save new code
	mov	ax,[OPrefix]		;Write old prefix code
	call	WriteCode
	pop	ax
	cmp	ax,[OMaxCode]		;Exceed code size?
	jb	CoBy6			;Less means no
	cmp	ax,HASHSIZE-1		;Currently less than 12 bits?
	jnb	CoBy5			;No. Write clear code
	inc	[OCodeSize]		;Increase number of bits
	shl	[OMaxCode],1		;Double max code size
	jmp	CoBy6			;Get next char
CoBy5:	mov	ax,[OClearCode] 	;Write a clear code
	call	WriteCode
	call	InitTable		;Reinit table
CoBy6:	mov	al,[OLastChar]		;Get last char and start over
CoBy7:	xor	ah,ah			;Turn char into code
CoBy8:	mov	[OPrefix],ax		;Set prefix code
	ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
WriteCode:
	mov	bx,WrBuffer		;Point into buffer offset of zero
	mov	di,[OBitOffset]
	push	di
	push	di
	mov	cl,3
	shr	di,cl			;Convert bit offset to byte offset
	pop	cx
	and	cx,7
	jz	WrCo1			;If zero, skip shift
	mov	dx,1
	shl	dx,cl
	mul	dx			;Rotate code
	mov	[bx+di+2],dl		;dl will catch bits rotated out
	or	al,[bx+di]		;Grab bits currently in buffer
WrCo1:	mov	[bx+di],ax		;Save data
	pop	ax			;Get OBitOffset
	add	ax,[OCodeSize]		;Adjust bit offset by code size
	mov	[OBitOffset],ax
	sub	ax,255*8		;Approaching end of buffer?
	jb	WrCoRet
	mov	[OBitOffset],ax
	mov	ax,255
	push	word [bx+255]		;Save last word to stack
	call	Flush			;Write the buffer
	pop	word [bx]		;Move last word to first position
WrCoRet:ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Param:   al = PaletteFlag
;Return:  CF is set on error
;Destroy: ax,cx,dx
GetVgaPalette:
	test	al,80h
	clc
	jz	SePaRet 		;Exit with clear CF
	and	al,07h			;al=BitsPerPixel-1
	xchg	ax,cx
	mov	ax,6
	shl	ax,cl			;ax=3*(1 Shl BitsPerPixel)
	call	CopyBlock		;Seek over color palette
SePaRet:ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Used registers:
;  si=EndBuffer
;  cx=Space counter
;  dl=Last char, dh=Eol flag
;  al=New char, ah=Roll
;Bit position formats:
;  1 - Count of spaces at EndOfLine
;  2 - Count of spaces between words
;  3 - One space=0, Two space=1
;  4 - Ascii=0, Space+Ascii=1
DoLstFile:
	call	ClrUse
	mov	si,EndBuffer
	xor	cx,cx			;Clear space counter
	mov	dx,0100h		;Set Eol flag and clear Last char
DoLst1: call	GetByte
	jbe	DoLstRe
	mov	ah,[Roll]
	cmp	al,0Ah
	je	DoLst4
	cmp	al,0Dh
	je	DoLst4
	cmp	al,20h
	je	DoLst3
	cmp	ah,4			;Is Ascii+Space mode?
	jz	DoLst5
	;Jcxz    DoLst8                  ;Spc2b - Add
	cmp	ah,1			;Is Eol mode?
	jz	DoLst6
	shr	dh,1			;Clear Eol flag
	jc	DoLst6
	jcxz	DoLst8			;Spc2b - Remove
	dec	cx
	jz	DoLst5
	cmp	ah,2			;Is Space between words?
	jz	DoLst5
	mov	cx,1
	jmp	DoLst5
DoLst3: inc	cx
	jmp	DoLst9
DoLst4: mov	dh,1			;Set Eol flag
	cmp	ah,1			;Is Eol mode?
	jnz	DoLst6			;Yes. Don't mess with last spaces
	cmp	dl,0Dh			;Keep 0Dh,0Ah together
	jz	DoLst6
DoLst5: mov	[si],cx
	push	ax
	mov	cl,1
	call	HandleByteCl
	pop	ax
	mov	cx,[si]
	test	dh,dh			;Is clear Eol flag?
	jnz	DoLst6
	inc	cx
DoLst6: jcxz	DoLst8
DoLst7: push	ax
	mov	al,20h
	call	PutByte
	pop	ax
	jc	DoLstRe
	loop	DoLst7
DoLst8: call	PutByte
	jc	DoLstRe
DoLst9: mov	dl,al
	call	IsAllDone
	jnz	DoLst1			;Loop back until all done
	clc
DoLstRe:ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Used registers:
;  si=Input byte position
;  di=BytesPerPlane
;  bp=PcxBytesPerLine
;  dh=Plane count, dl=Plane counter
DoPcxFile:
	mov	cx,sizeof.PcxHeaderType
	call	ReadHeader
	jc	PcxErr
	cmp	word [PcxHeader],050Ah	;Check if Id = 0Ah
	jnz	PcxErr			;and Version = 5
	call	WriteHeader
	jnc	DoPcx1
PcxErr: stc
	ret
DoPcx1: mov	ax,[PcxHeader.PcxXMax]
	sub	ax,[PcxHeader.PcxXMin]
	inc	ax
	xor	cx,cx				;Routine to handle
	mov	cl,[PcxHeader.PcxBitsPerPixel]	;   1 BitPerPixel
	mul	cx				; * 4 ColorPlanes
	mov	cl,8				;
	cmp	cx,dx				;Avoid divide by 0
	jbe	PcxErr				;Divide by byte size
	div	cx				;ax=BytesPerPlane
	xchg	di,ax
	mov	bp,[PcxHeader.PcxBytesPerLine]
	mov	cx,[PcxHeader.PcxYMax]
	sub	cx,[PcxHeader.PcxYMin]
	inc	cx
DoPcx2: push	cx
	mov	dh,[PcxHeader.PcxColorPlanes]
	mov	si,EndBuffer
	push	si
	call	ReadPcxLine
	pop	si
	jc	DoPcx5
	mov	dl,dh
DoPcx3: mov	cx,di			;BytesPerPlane
	push	si
DoPcx4: push	cx
	call	HandleByte
	pop	cx
	inc	si
	loop	DoPcx4
	pop	si
	add	si,bp
	dec	dl
	jnz	DoPcx3
	call	WritePcxLine
DoPcx5: pop	cx			;PicDepth
	jc	PcxErr
	call	IsAllDone
	loopne	DoPcx2			;Loop back until all done
	call	IsSeek
	jz	DoPcx7
	mov	dl,[PcxHeader.PcxBitsPerPixel]
	cmp	dx,0108h		;Is ColorPlane=1 and BitsPerPixel=8?
	jnz	DoPcx7
	mov	ax,769
	call	CopyBlock		;Seek over color palette
	jc	PcxErr
DoPcx7: clc
	ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Used registers:
;  si=Color plane start offset
;  cx,bp=PcxBytesPerLine
;  dh=Plane count, dl=Plane counter
ReadPcxLine:
	mov	dl,dh
RePcx1: mov	cx,bp
RePcx2: mov	ax,[RLECountW]		;Get old counter and color
	test	ah,ah			;Is counter = 0?
	jnz	RePcx3
	call	GetByte 		;Get counter/color
	jbe	RePcxEr
	mov	ah,1
	cmp	al,0C0h
	jb	RePcx3
	and	al,3Fh
	mov	bh,al
	call	GetByte 		;Get color
	jbe	RePcxEr
	mov	ah,bh
RePcx3: mov	[si],al
	inc	si
	dec	ah
	mov	[RLECountW],ax
	loop	RePcx2
	dec	dl
	jnz	RePcx1
	clc
	ret
RePcxEr:stc
	ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Used registers:
;  si=Color plane start offset
;  bx=Pixel position in color plane
;  bp=PcxBytesPerLine
;  dh=Plane count, dl=Plane counter
;  al=Counter+0C0h, ah=Color
WritePcxLine:
	call	IsSeek
	jz	WrPcx8
	mov	dl,dh
	mov	si,EndBuffer
WrPcx1: xor	bx,bx
WrPcx2: mov	al,0C0h
	mov	ah,[bx+si]
WrPcx3: inc	bx
	inc	ax
	cmp	ah,[bx+si]
	jnz	WrPcx4
	cmp	al,0FFh
	jz	WrPcx5
	cmp	bx,bp
	jb	WrPcx3
WrPcx4: cmp	al,0C1h
	jnz	WrPcx5
	cmp	ah,0C0h
	jb	WrPcx6
WrPcx5: call	PutByte 		;Put counter
	jc	WrPcxRe
WrPcx6: mov	al,ah
	call	PutByte 		;Put color
	jc	WrPcxRe
WrPcx7: cmp	bx,bp
	jb	WrPcx2
	add	si,bp
	dec	dl
	jnz	WrPcx1
WrPcx8: clc
WrPcxRe:ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Used registers:
;  si=Input byte position
;  cx=TgaLine / LinePos / BitPosition
;  bp=BytesPerLine
DoTgaFile:
	mov	cx,sizeof.TgaHeaderType
	call	ReadHeader
	jc	TgaErr
	mov	al,0F8h
	and	al,[TgaHeader.TgaDataType]	;Is compressed?
	jz	DoTga0
	and	[TgaHeader.TgaDataType],7	;Clear compressed flag
DoTga0: mov	[IsCompressed],al
	call	WriteHeader
	jnc	DoTga1
TgaErr: stc
	ret
DoTga1: mov	al,[TgaHeader.TgaBitsPerPixel]
	mov	cl,3
	ror	al,cl
	cmp	al,4			;Check BitsPerPixel in [8,16,24,32]
	ja	TgaErr
	cbw
	mov	[BytesPerPixel],ax
	mov	cx,ax
	jcxz	TgaErr
	mul	[TgaHeader.TgaX]
	xchg	bp,ax
	loop	DoTga2			;Only 8b color may have palette
	mov	ax,3
	mul	[TgaHeader.TgaMapLength]
	call	CopyBlock		;Seek over color palette
	jc	TgaErr
DoTga2: mov	cx,[TgaHeader.TgaY]
DoTga3: push	cx
	mov	cx,bp
	jcxz	DoTga8
	call	ReadTgaLine
	jc	DoTga8
	mov	dx,bp
DoTga4: mov	cl,[Roll]
	cmp	[TgaHeader.TgaBitsPerPixel],16 ;16b color need some extra work
	jnz	DoTga6
	cmp	cl,8			;Special code 8 = bit 16 of word
	jz	DoTga5
	call	HandleByteCl
	add	cl,5
	call	HandleByteCl
	sub	cl,3
DoTga5: inc	si
	dec	dx
	jz	DoTga7
DoTga6: call	HandleByteCl
	inc	si
	dec	dx
	jnz	DoTga4
DoTga7: mov	dx,EndBuffer
	mov	cx,bp
	call	PutBlock
DoTga8: pop	cx			;PicDepth
	jc	TgaErr
	call	IsAllDone
	loopne	DoTga3			;Loop back until all done
	clc
	ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Used registers:
;  si=EndBuffer
;  bx=Pixel position in current line
;  bp=BytesPerLine
;  cx=BytesPerPixel
;  dl=Pixel counter, dh=Pixel count
ReadTgaLine:
	xor	bx,bx
	mov	si,EndBuffer
	cmp	[IsCompressed],bl
	jz	ReTga5			;cx=BytesPerLine
ReTga1: call	GetByte 		;Get counter
	jc	ReTgaRe
	mov	cx,[BytesPerPixel]
	test	al,80h			;Is high bit on?
	jz	ReTga4			;No. Copy bytes
	mov	dh,al			;Yes. Repeat next color
ReTga2: call	GetByte 		;Get color
	jc	ReTgaRe
	push	bx
	mov	dl,dh
ReTga3: mov	[bx+si],al
	mov	di,bx
	add	bx,[BytesPerPixel]
	dec	dl
	js	ReTga3			;Loop back while >7Fh
	pop	bx
	inc	bx
	loop	ReTga2
	mov	bx,di
	inc	bx
	jmp	ReTga9
ReTga4: inc	ax
	mul	cl
	xchg	ax,cx
ReTga5: call	GetByte 		;Get color
	jc	ReTgaRe
	mov	[bx+si],al
	inc	bx
	loop	ReTga5
ReTga9: cmp	bx,bp
	jb	ReTga1
	clc
ReTgaRe:ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
DoTxtFile:
	mov	si,EndBuffer
	xor	di,di			;Clear line length
	mov	ch,1			;First Ascii flag for option '2'
	cmp	[UseFlag],','
	jz	DoTxt1
	call	ClrRollCount
DoTxt1: call	GetByte
	jbe	DoTxtRe
	cmp	[UseFlag],'='
	jnz	DoTxt6
	push	ax
	cmp	al,0Ah
	jz	DoTxt4
	inc	di
	cmp	al,0Dh
	jnz	DoTxt5
	call	IsHide
	jnz	DoTxt5
DoTxt2: call	IsHideSize0
	jz	DoTxt5
	mov	ax,[LineLen]
	test	ax,ax
	jz	DoTxt5
	cmp	ax,di
	jb	DoTxt5
	inc	di
	mov	al,[UseMin]
	call	PutTxtByte
	jnc	DoTxt2
	pop	ax
	ret
DoTxt4: xor	di,di
DoTxt5: pop	ax
DoTxt6: call	PutTxtByte
	jc	DoTxtRe
	call	IsAllDone
	jnz	DoTxt1			;Loop back until all done
	clc
DoTxtRe:ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
DoVocFile:
	mov	cx,sizeof.VocHeaderType
	call	ReadHeader
	jc	VocErr
	cmp	word [VocHeader.VocID],'Cr'
	jnz	VocErr
	call	WriteHeader
	jnc	DoVoc1
VocErr: stc
	ret
DoVoc1: mov	ax,[VocHeader.VocOfsData]
	sub	ax,sizeof.VocHeaderType
	jz	DoVoc3
	xor	dx,dx
DoVoc2: call	CopyBlockBig
	jc	VocErr
DoVoc3: call	CopyByte		;Get type byte
	jbe	VocErr
	cmp	al,00h			;00h = Terminator
	jz	DoVoc8
	xchg	ax,di			;Save type byte
	call	CopyByte		;Get 3-byte block size
	mov	dl,al
	call	CopyByte
	mov	dh,al
	call	CopyByte
	xor	ah,ah
	xchg	ax,dx			;Swap High word / Low word
	xchg	ax,di			;Swap Low word / type byte
	cmp	al,2
	jb	DoVoc4			;01h = Sound data
	jz	DoVoc5			;02h = Sound continue
	xchg	ax,di			;Else copy block
	jmp	DoVoc2
DoVoc4: call	CopyByte		;Copy sample rate
	call	CopyByte		;Copy compression type
	cmp	al,0			;Is compression type 8-bits = 0?
	jnz	VocErr			;No. Error exit
	sub	di,2
	sbb	dx,0
	jc	DoVoc3
DoVoc5: call	HandleBlock
	jc	DoVocRe
	call	IsAllDone
	jnz	DoVoc3			;Loop back until all done
DoVoc8: clc
DoVocRe:ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
DoWavFile:
	mov	cx,sizeof.WavHeaderType
	call	ReadHeader
	jc	WavErr
	cmp	word [WavHeader.WavFileID.RiffDataID],'WA'
	jnz	WavErr
	cmp	word [WavHeader.WavDataID],'da'
	jnz	WavErr
	cmp	[WavHeader.WavFormTag],1	;Uncompressed sample
	jnz	WavErr
	call	WriteHeader
	jnc	DoWav1
WavErr: stc
	ret
DoWav1: cmp	[WavHeader.WavFormatSpecific],8 ;Format specific data area
	jbe	DoWav2
	mov	[DataSize],2
DoWav2: mov	di,word [WavHeader.WavDataLen]
	mov	dx,word [WavHeader.WavDataLen+2]
	call	HandleBlock
	jc	WavErr
	call	IsAllDone
	jz	DoWav8			;Loop back until all done
DoWav7: call	CopyByte
	jc	WavErr
	jnz	DoWav7
DoWav8: clc
	ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
DoRawFile:
	mov	di,word [HostSize]
	mov	dx,word [HostSize+2]	;Continue to HandleBlock
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Used registers: ax,bx,cx,si
HandleBlock:
	mov	si,EndBuffer
	xor	ch,ch
HaBl1:	sub	di,1
	sbb	dx,0
	jc	HaBlOk
	call	GetByte
	jbe	HaBlEr
	test	ch,ch
	jnz	HaBl2
	mov	[si],al
	call	HandleByte
	mov	al,[si]
HaBl2:	call	PutByte
	jc	HaBlEr
	inc	ch
	cmp	ch,[DataSize]
	jb	HaBl1
	call	IsAllDone
	jnz	HandleBlock		;Loop back until all done
HaBlOk: clc
	ret
HaBlEr: stc
	ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
HandleSwap:
	cmp	[si],bl
	jz	HaSw1
	cmp	[si],bh
	jnz	HaSw3
	stc
HaSw1:	rcl	byte [si],1
	call	ClrRollCount
	mov	cl,1
	call	HaBy1
	shr	byte [si],1
	mov	ax,[UseMinMaxW]
	jnc	HaSw2
	mov	al,ah
HaSw2:	mov	[si],al
HaSw3:	ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Used registers: ax,bx,cl
HandleByte:
	mov	cl,[Roll]
HandleByteCl:
	mov	bx,[UseMinMaxW]
	mov	al,[UseFlag]
	cmp	al,'='
	jz	HandleSwap
	cmp	al,','
	jnz	HaBy1
	call	IsInRange
	jnc	HaByRet
HaBy1:	call	IsDummy
	jc	HaByRet
	push	cx
	mov	ch,[RollCount]
HaBy2:	call	IsSeek
	jz	HaBy6
	shl	[HideCount],1		;Is HideBits empty?
	jc	HaBy6			;No. Scroll bits
	call	IsHeadDone
	mov	al,[bx] 		;Next header byte
	jnz	HaBy4
HaBy3:	call	IsHideSize0		;If HideSize=0 nothing to do
	jz	HaBy9
	call	GetHideBits
HaBy4:	mov	[HideBits],al
	mov	[HideCount],0FEh
HaBy6:	rcr	byte [si],cl		;Scroll data byte down
RclPos: rcr	[HideBits],1		;Get/Put bit (selfmodify code)
	rcl	byte [si],cl		;Scroll data byte up
	call	IsSeek
	jnz	HaBy8
	add	[HideCount],20h 	;Is all 8 bit in HideBits / Clear
	jnc	HaBy8
	call	IsHeadDone
	jz	HaBy7
	mov	al,[HideBits]
	mov	[bx],al
	jmp	HaBy8
HaBy7:	call	IsHideSize0		;If HideSize=0 nothing to do
	jz	HaBy8
	call	PutHideBits		;Put new byte and return
HaBy8:	add	cx,0FF01h
	jc	HaBy2
HaBy9:	pop	cx
HaByRet:ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Used registers: ax,bx
PutHideBits:
	mov	bx,[IdeaPos]
	mov	al,[HideBits]
	mov	[bx+IdeaBuffer],al
	call	GeHiBi2
	cmp	bx,BUFFERSIZE-1 	;Is buffer full?
	jb	PuHiRet
FlushIdeaBuffer:			;Not pretty, but save some bytes
	mov	ax,[IdeaPos]
	call	EncryptBuffer
	mov	ah,40h			;Dos WriteFile
RWIdeaBuffer:
	push	cx
	push	dx
	xor	cx,cx
	xchg	cx,[IdeaPos]		;Get byte count with clear
	mov	dx,IdeaBuffer
	call	DoHideH 		;Read/Write encrypted buffer
	pop	dx
	pop	cx
PuHiRet:ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Used registers: ax,bx
GetHideBits:
	mov	al,0
	call	IsHide
	jnz	GeHiBi2 		;Jump if clear mode
	mov	ax,[IdeaPos]
	cmp	ax,[IdeaCount]
	jb	GeHiBi1
	mov	[IdeaPos],BUFFERSIZE	;Read up to 2k into buffer
	mov	ah,3Fh			;Dos ReadFile
	call	RWIdeaBuffer
	mov	[IdeaCount],ax
	call	EncryptBuffer
GeHiBi1:mov	bx,[IdeaPos]
	mov	al,[bx+IdeaBuffer]
GeHiBi2:inc	[IdeaPos]
	sub	word [HideSize],1	;Decrement HideSize
	sbb	word [HideSize+2],0
	ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Return:  bx = Byte position in HideSize (ZF is set if hide size already done)
IsHeadDone:
	xor	bx,bx
	sub	bl,[HeadSize]
	jz	GeHeRet
	add	bl,[WhatLength]
	dec	[HeadSize]
	add	bx,HideSize
GeHeRet:ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Used registers: ax,dx
;Note: Can't write back because header may change (TGA)
ReadHeader:
	mov	dx,Header		;Start of host file header
	call	GetBlock
	jc	ReHeRet
	mov	[HeaderSize],ax
	cmp	ax,cx
ReHeRet:ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Return:  al = Byte from InputBuffer (ZF is set if none)
;         CF is set on error
GetByte:
	push	bx
	mov	bx,[InBufPos]
	cmp	bx,[InBufCount]
	jb	GeBy1
	push	cx
	push	dx
	mov	cx,BUFFERSIZE
	mov	dx,InputBuffer
	mov	ah,3Fh			;Dos ReadFile
	call	DoHostH
	mov	[InBufCount],ax
	pop	dx
	pop	cx
	jc	GeByEx
	and	ax,ax
	jz	GeByClc
	xor	bx,bx
GeBy1:	mov	al,[bx+InputBuffer]
	inc	bx
	mov	[InBufPos],bx
GeByClc:clc
GeByEx: pop	bx
	ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Param:   cx = Count of bytes to read
;         ds:dx -> Buffer
;Return:  ax = Number of bytes actually read (ZF is set if none)
;         CF is set on error (dx=Error Asciiz)
GetBlock:
	push	bx
	push	cx
	mov	bx,dx
	jcxz	GeBl3
GeBl1:	call	GetByte
	jc	GeBlEx
	jz	GeBl3
	mov	[bx],al
	inc	bx
	loop	GeBl1
GeBl3:	xchg	ax,bx
	sub	ax,dx
GeBlEx: pop	cx
	pop	bx
	ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
PutTxtByte:
	mov	[si],al
	mov	bx,[UseMinMaxW]
	mov	al,[UseFlag]
	cmp	al,'='
	jz	PuTx4
	cmp	al,','
	jnz	PuTx1
	call	IsInRange
	jnc	PuTx5
	call	HandleByte
	jmp	PuTx6
PuTx1:	mov	al,[si]
	and	al,0DFh 		;Convert to uppercase
	cmp	al,'A'			;Test if Ascii char
	jb	PuTx5
	cmp	al,'Z'
	ja	PuTx5
PuTx2:	mov	al,[Roll]
	cmp	al,3			;Is NOT first Ascii mode?
	ja	PuTx3
	sub	al,2			;Is first Ascii mode?
	jb	PuTx3			;2->0, 3->1
	shr	ch,1
	sbb	al,0			;Swap with Ascii flag
	jz	PuTx6			;Jump over if wrong flag
PuTx3:	mov	cl,6			;Use bit 6 for UPPER/lower case
	call	HandleByteCl
	jmp	PuTx6
PuTx4:	cmp	[si],bl
	jz	PuTx2
	cmp	[si],bh
	jz	PuTx2
PuTx5:	mov	ch,1			;First Ascii flag
PuTx6:	mov	al,[si]
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Param:   al = Byte to put in OutputBuffer
PutByte:
	call	IsSeek
	jz	PuByClc
	push	bx
	mov	bx,[OutBufCount]
	mov	[bx+OutputBuffer],al
	inc	[OutBufCount]
	cmp	bx,BUFFERSIZE-1 	;Is buffer full?
	pop	bx
	jb	PuByClc
	call	FlushTempH
	jc	PuByRet
PuByClc:clc
PuByRet:ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Return:  al = Byte read (ZF is set if no bytes read)
;         CF is set on error
CopyByte:
	call	GetByte
	jbe	CoByRet
	pushf
	call	PutByte
	jc	CoByErr
	popf
	ret
CoByErr:popf
	stc
CoByRet:ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
WriteHeader:
	mov	dx,Header		;Start of host file header
	mov	cx,[HeaderSize]
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Param:   cx = Count of bytes to write
;         ds:dx -> Buffer
;Return:  CF is set on error
PutBlock:
	push	bx
	push	cx
	call	IsSeek
	jz	PuBlClc
	mov	bx,dx
	jcxz	PuBlClc
PuBl1:	mov	al,[bx]
	call	PutByte
	jc	PuBlEx
	inc	bx
	loop	PuBl1
PuBlClc:clc
PuBlEx: pop	cx
	pop	bx
	ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Return:  CF is set on error
FlushTempH:
	push	ax
	push	bx
	push	cx
	push	dx
	xor	cx,cx
	xchg	cx,[OutBufCount]	;Get byte count with clear
	mov	bx,[TempHandle]
	mov	dx,OutputBuffer
	mov	ah,40h			;Dos WriteFile
	int	21h
	jc	FlTe1
	cmp	ax,cx			;Is all bytes written
FlTe1:	pop	dx
	pop	cx
	pop	bx
	pop	ax
FlTeRet:ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Param:   dx:ax -> Count of bytes to copy
;Return:  CF is set on error
;Destroy: ax,cx,dx
CopyBlock:
	xor	dx,dx
CopyBlockBig:
	mov	cx,ax
CoBl1:	sub	cx,1
	sbb	dx,0
	jc	CoBlClc
	call	CopyByte
	ja	CoBl1
	stc
	ret
CoBlClc:clc
	ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Return:  ZF is set if HideSize = 0
;Destroy: ax
IsHideSize0:
	mov	ax,word [HideSize+2]
	or	ax,word [HideSize]
	ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Return:  ZF is set if all done
;Destroy: ax
IsAllDone:
	call	IsHideSize0
	or	ax,[HeadSizeW]		;HeadSize and HideCount
	jnz	IsAlRet
IsSeek: cmp	[DecodeMode],'-'
IsAlRet:ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
IsHide: cmp	[DecodeMode],'+'
	ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Return:  CF is set if dummy bit
;Destroy: ax
IsDummy:
	cmp	[DummyCounter],0
	jz	IsDu2
	dec	[DummyCounter]
	jnz	IsDu1
	mov	ax,[DefCount]
	mov	[Position],ax
IsDu1:	stc
	ret
IsDu2:	dec	[Position]
	jnz	IsDu3
	mov	ax,[DefDummyCount]
	mov	[DummyCounter],ax
IsDu3:	clc
	ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
OpenFiles:
	mov	al,[UseFlag]
TeRa0:	cmp	al,','
	jnz	TeRa1
	mov	cx,[RollW]
	mov	al,0FEh
	shl	al,cl
	mov	cl,ch
	dec	cx
	rol	al,cl
	and	[UseMin],al
	not	al
	or	[UseMax],al
TeRa1:	mov	ax,[DefCount]
	mov	[Position],ax
	xor	ax,ax
	mov	[InBufCount],ax
	mov	[OutBufCount],ax
	mov	[IdeaCount],ax
	mov	[IdeaPos],ax
	mov	[RLECountW],ax
	mov	[FileErr],al		;Clear error flag
	cmp	[WhatLength],al
	jnz	OpFi0
	dec	ax			;Set maximum HideSize for 'X' seek
OpFi0:	mov	word [HideSize],ax
	mov	word [HideSize+2],ax
	mov	dx,[HostName]
	mov	ax,3D00h		;Dos OpenFile (ReadOnly)
	int	21h
	jc	OpFiErr
	mov	[HostHandle],ax 	;Save Host file handle
	xchg	ax,bx
	mov	ax,5700h		;Dos GetFTime
	int	21h
	mov	[FTime],cx
	mov	[FDate],dx
	mov	ax,4202h		;Dos SeekFile (EndOfFile)
	cwd
	xor	cx,cx
	int	21h
	mov	word [HostSize],ax
	mov	word [HostSize+2],dx
	mov	ax,4200h		;Dos SeekFile (StartOfFile)
	cwd
	xor	cx,cx
	int	21h
	call	OpenHideFile
	jc	OpFiCl1 		;Close Host file
	call	IsSeek
	jz	OpFi1
	mov	dx,TempName
	xor	cx,cx
	mov	ah,3Ch			;Dos CreateFile
	int	21h
	mov	[TempHandle],ax 	;Save Temp file handle
	jc	OpFiCl2 		;Close Hide and Host files
OpFi1:	clc
	ret
OpFiCl2:call	CloseHide
OpFiCl1:call	CloseHost
OpFiErr:stc
	ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
CloseFiles:
	call	CloseHide
	call	CloseHost
	call	IsSeek
	jz	ClFiRet
	call	FlushTempH
	jc	ClFiRet
	mov	cx,[FTime]
	mov	dx,[FDate]
	mov	bx,[TempHandle]
	mov	ax,5701h		;Dos SetFTime
	int	21h
	mov	ah,3Eh			;Dos CloseFile (Temp file)
	int	21h
	cmp	[FileErr],None
	jnz	ClFiRet
	mov	dx,[HostName]
	call	IsHide
	jz	ClFi2
	call	OverWriteFile		;Only overwrite if clear mode
ClFi2:	mov	ah,41h			;Dos DeleteFile
	int	21h			;CF is set if try to delete in CD
	mov	di,[PureHostName]
	mov	dx,TempName
	mov	ah,56h			;Dos RenameFile
	int	21h
ClFiRet:ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
OpenHideFile:
	mov	dx,[HideName]
	mov	ax,3D00h		;Dos OpenFile (ReadOnly)
	call	IsHide
	jz	OpHiFi1
	call	IsSeek
	jnz	OpHiFi4
	cmp	[WriteMode],'A'
	jnz	OpHiFi0
	mov	ax,3D02h		;Dos OpenFile (ReadWrite)
	int	21h
	xchg	ax,bx
	mov	ax,4202h		;Dos SeekFile (EndOfFile)
	jnc	OpHiFi3
OpHiFi0:xor	cx,cx
	mov	ah,3Ch			;Dos CreateFile
OpHiFi1:int	21h
	jc	OpHiRet
	xchg	ax,bx
	call	IsSeek
	jz	OpHiFi4
	mov	ax,4202h		;Dos SeekFile (EndOfFile)
	cwd
	xor	cx,cx
	int	21h
	mov	word [HideSize],ax
	mov	word [HideSize+2],dx
	mov	ax,4200h		;Dos SeekFile (StartOfFile)
OpHiFi3:cwd
	xor	cx,cx
	int	21h
OpHiFi4:mov	[HideHandle],bx 	;Save HideHandle
	xor	ax,ax
	mov	al,[WhatLength]
	mov	[HeadSizeW],ax		;HeadSize and HideCount
	clc
OpHiRet:ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
CloseHide:
	call	IsHide
	jz	ClHi1
	call	IsSeek
	jnz	ClHiRet
	call	FlushIdeaBuffer
ClHi1:	mov	ah,3Eh			;Dos CloseFile
DoHideH:mov	bx,[HideHandle]
	int	21h
ClHiRet:ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
CloseHost:
	mov	ah,3Eh			;Dos CloseFile
DoHostH:mov	bx,[HostHandle]
	int	21h
	ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
OverWriteFile:
	push	dx
	mov	ax,3D01h		;Dos OpenFile (WriteOnly)
	int	21h
	jc	OvWrErr
	xchg	ax,bx
	mov	di,word [HostSize]
	mov	si,word [HostSize+2]
OvWr1:	mov	cx,Roll - TxtInfo
	sub	di,cx
	sbb	si,0
	jnc	OvWr2
	add	cx,di
	jz	OvWr3
	xor	di,di
	xor	si,si
OvWr2:	mov	dx,TxtInfo
	mov	ah,40h			;Dos WriteFile
	int	21h
	jnc	OvWr1
OvWr3:	mov	ah,3Eh			;Dos CloseFile
	int	21h
OvWrErr:pop	dx
	ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
GetIDEAKey:
	mov	dx,EnterKey
	mov	ax,0900h
	mov	di,Passphrase-1
	mov	cx,IdeaBuffer-Passphrase+BUFFERSIZE+1
	rep	stosb		; Overwrite data area
	int	21h		; Print message requesting key
	mov	dx,Password
	mov	ah,0Ah
	int	21h		; Get passphrase from user
	mov	dx,TxtCrLf
	mov	ah,09h
	int	21h		; Print new line
	mov	cx,PassLen/8	; Hash passphrase down to 128 bits
	mov	si,Passphrase
TandemDM:
	push	cx		; push block counter    <0
	push	si		; push data             <1
	mov	cl,4
	mov	si,Key+8	; -- Let first half of key = second half of hash (G)
	mov	di,hashKey
	rep	movsw		; after: di->key+8
	pop	si		; si->data              1>
	push	si		; push data             <1
	call	Convert 	; -- Let second half of key = data
	mov	di,hashKey
	push	di		; push key              <2
	call	ExpandKey	; -- Expand key : Key = {G, data}
	mov	cl,4
	mov	si,Key
	mov	di,w_val
	push	di		; push w                <3
	rep	movsw		; -- Let W = first half of hash (H)
	pop	di		; di->w                 3>
	pop	si		; si->key               2>
	push	si		; push key              <2
	call	IDEA		; -- W = IDEA encrypt (W, {G,data})
	mov	si,di		; si->w
	mov	di,Key
	call	DoGHsub 	; -- Update H with (H xor W)
	pop	di		; di->key               2>
	pop	si		; si->data              1>
	push	si		; push data             <1
	push	di		; push key              <2
	call	Convert 	; -- Let first half of key = data
	mov	cl,4
	mov	si,w_val	; di->key+8
	rep	movsw		; -- Let second half of key = W
	pop	di		; di->key               2>
	call	ExpandKey	; -- Expand key : Key = {data, W}
	mov	cl,4
	mov	si,Key+8
	mov	di,g0
	push	si		; push g                <2
	push	di		; push g0               <3
	rep	movsw		; -- Let G0 = G
	pop	di		; di->g0                3>
	mov	si,hashKey
	call	IDEA		; -- G0 = IDEA encrypt (G0, {data, W})
	mov	si,di		; si->g0
	pop	di		; di->g                 2>
	call	DoGHsub 	; -- Update G with (G xor G0)
	pop	si		; si->data              1>
	add	si,8		; data+=8
	pop	cx		; cx=block counter      0>
	loop	TandemDM	; -- Continue hashing until no more blocks
	mov	di,Key		; Expand hashed passphrase to IDEA key
; Expand user key to IDEA encryption key
; Entry:  si -> userkey, di -> buffer for IDEA key (can equal si)
; Exit:   di -> IDEA key
ExpandKey:
	add	di,16
	mov	bl,8
Rotate: mov	ax,bx			; Determine which two of the previous
	and	al,7			;   eight words are needed for this
	cmp	al,6			;   key expansion round
	mov	ax,[di-14]
	mov	dx,[di-12]
	jb	Update
	mov	dx,[di-28]
	jz	Update
	mov	ax,[di-30]
Update: mov	cl,9
	shl	ax,cl
	mov	cl,7
	shr	dx,cl			; Calculate the rotated value
	or	ax,dx
	stosw				; and save it
	inc	bx
	cmp	bl,52
	jnz	Rotate			; Continue until 52 words updated
	ret

EncryptBuffer:
	cmp	[IsIdea],0
	jz	EnBufRe
	push	bx
	push	cx
	push	dx
	push	si
	push	di
	push	bp
	mov	si,IdeaBuffer		; si -> data (start)
	add	ax,7
	mov	cl,3
	shr	ax,cl
	xchg	cx,ax			; cx = number of blocks to encrypt
Block:	push	cx
	push	si			; Save current position
	mov	si,Key
	mov	di,CFBBuffer		; Encrypt 8 byte CFB buffer (di)
	call	IDEA
	mov	cx,4			; Process 4 words
	pop	si			; si -> data, di -> CFB buffer

	call	IsSeek
	jz	Decrypt

; Cipher Feedback

Encrypt:lodsw				; Get a word from file buffer
	xchg	ah,al			;   and convert it to little-endian
	xor	ax,[di] 		; XOR data with CFB buffer
	stosw				; Replace ciphertext in CFB buffer
	xchg	ah,al			; Convert back to big-endian
	mov	[si-2],ax		;   and store in file buffer
	loop	Encrypt
	jmp	DoNextBlock		; Skip over Decrypt routine

Decrypt:lodsw				; Get word from file buffer
	xchg	ah,al			;   convert it to little-endian
	mov	dx,[di] 		; Get word from CFB buffer
	stosw				; Update CFB buffer
	xor	ax,dx			; XOR data with CFB buffer
	xchg	ah,al			; Convert back to big-endian
	mov	[si-2],ax		;   and store in file buffer
	loop	Decrypt

DoNextBlock:
	pop	cx
	loop	Block			; Continue until all blocks processed
	pop	bp
	pop	di
	pop	si
	pop	dx
	pop	cx
	pop	bx
EnBufRe:ret
; Buffer Encrypted

; Convert string to little-endian words
; (called by Tandem DM hashing routine)
Convert:mov	cl,4
Conv2:	lodsw
	xchg	ah,al
	stosw
	loop	Conv2
	ret

; XOR update 64-bit buffer
; (called by Tandem DM hashing routine)
DoGHsub:mov	cx,4
DoGH2:	lodsw
	xor	ax,[di]
	stosw
	loop	DoGH2
	ret

; IDEA subroutine
; Entry:  si -> key, di -> input data
; Exit:   di -> output data, all other registers trashed
; Refer to the PGP IDEA source for a better explanation
; of the algorithm and the optimisations
; Thanks to Bill Couture <bcouture@cris.com> for speed optimisations

x0	equ	bx
x1	equ	cx
x2	equ	bp
x3	equ	di

IDEA:	push	di
	mov	dx,[di]
	mov	x1,[di+2]
	mov	x2,[di+4]
	mov	x3,[di+6]		; note that di is over-written last
	mov	ax,8			; Eight rounds
Round:	push	ax
	call	MulMod
	xchg	x0,ax			; x0 *= *key++
	lodsw
	add	x1,ax			; x1 += *key++
	lodsw
	add	x2,ax			; x2 += *key++
	mov	dx,x3
	call	MulMod
	xchg	x3,ax			; x3 *= *key++
	push	x1			; s0 = x1
	push	x2			; s1 = x2
	xor	x2,x0			; x2 ^= x0
	xor	x1,x3			; x1 ^= x3
	mov	dx,x2
	call	MulMod
	add	x1,ax			; x2 *= *key++
	xchg	x2,ax			; x1 += x2
	mov	dx,x1
	call	MulMod
	add	x2,ax			; x1 *= *key++
	xor	x0,ax			; x0 ^= x1
	xor	x3,x2			; x3 ^= x2
	pop	x1
	pop	dx
	xor	x1,ax			; x1 ^= s1
	xor	x2,dx			; x2 ^= s0
	mov	dx,x0
	pop	ax
	dec	ax			; Continue until no more rounds
	jnz	Round
	call	MulMod
	mov	dx,x3
	pop	di
	push	di
	stosw				; save x0 *= *key++
	lodsw
	add	ax,x2			; x2 += *key++
	stosw
	lodsw
	add	ax,x1			; x1 += *key++
	stosw
	call	MulMod			; x3 *= *key++
	stosw
	pop	di
	ret

; Multiplication modulo 65537
; ax = [si] * dx
MulMod: push	dx
	lodsw
	mul	dx
	sub	ax,dx
	pop	dx
	jnz	NotZero
	inc	ax
	sub	ax,[si-2]
	sub	ax,dx
	ret
NotZero:adc	ax,0
	ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Collecting all data stuff
TempName	db  "Stega.Tmp",0
TxtUnknownForm	db  07h,"Unknown format",0Dh,0Ah,"$"
TxtOpenError	db  07h,"Can't open file",0Dh,0Ah,"$"
TxtTooBig	db  07h,"Hide file too big",0Dh,0Ah,"$"
TxtFileError	db  07h,"File error",0Dh,0Ah,"$"
TxtInfo 	db  "STEGA v2.01 - Tiny steganography program"
		db  " by Asko Vuori, Finland",0Dh,0Ah
		db  "Formats: Bmp/Gif/Lst/Pcx/Raw/Tga/Txt/Voc/Wav",0Dh,0Ah
		db  "Usage:",0Dh,0Ah
		db  "  STEGA +|-[ABs,cHtLwPs,b,dRSUs,e] Host.Gif Hide.Ext"
TxtCrLf 	db  0Dh,0Ah,"$"
WriteMode	db  "O" 	;"A"=Append, "O"=Overwrite
UseFlag 	db  "x" 	;","=Use range, "="=2 bytes, "x"=None
DataSize	db  1
RollW		= $
RollCount	db  0		;Bit count in same byte 0 to 7 (default 0)
Roll		db  1		;Bit position 1 to 8 (default 1)
IsIdea		db  0		;Is IDEA scramble (default 0)
WhatLength	db  4		;0=no length, 2=16-bit, 4=32-bit
DummyCounter	dw  0		;Start at first byte
DefCount	dw  1		;Use bit count (default 1)
DefDummyCount	dw  0		;Dummy bit count (default 0)
LineLen 	dw  0

EnterKey	db  "Enter key: $"
Password	db  PassLen,?
Passphrase	db  PassLen dup(?)
		db  ?		;Comment out if assembly inserts NOP
align 2
; Data used by IDEA routine
IdeaCount	dw  ?		;Bytes in IdeaBuffer
IdeaPos 	dw  ?		;Position in IdeaBuffer
; Data used by Tandem DM hashing routine
Key		dw  8		dup(?)
w_val		dw  4		dup(?)
g0		dw  4		dup(?)
hashKey 	dw  52		dup(?)
; Data used by CFB routine
CFBBuffer	db  8		dup(?)
IdeaBuffer	db  BUFFERSIZE	dup(?)

; Variables outside program
FileErr 	db  ?		;File errors
DecodeMode	db  ?		;'+'=Hide, '-'=Seek, '#'=Clear
HideBits	db  ?
IsCompressed	db  ?		;0=Not compressed, else compressed
UseMinMaxW	= $
UseMin		db  ?
UseMax		db  ?
HeadSizeW	= $
HeadSize	db  ?		;2 downto 0 if 16-bit, 4 downto 0 if 32-bit
HideCount	db  ?		;Hide=Shl bits, Seek=Add 20h
HideHandle	dw  ?		;Hide file handle
HostHandle	dw  ?		;Host file handle
TempHandle	dw  ?		;Temporary file handle
HideName	dw  ?		;Offset of hide file name
HostName	dw  ?		;Offset of host file name
PureHostName	dw  ?		;Host file name without path
HostSize	dd  ?		;32-bit Host file length
HideSize	dd  ?		;32-bit Hide file length
FTime		dw  ?		;Host file time/date
FDate		dw  ?
HeaderSize	dw  ?		;Bytes in file Header
InBufPos	dw  ?		;Position in InputBuffer
InBufCount	dw  ?		;Bytes in InputBuffer
OutBufCount	dw  ?		;Bytes in OutputBuffer
BytesPerPixel	dw  ?
RLECountW	= $
RLEColor	db  ?		;Save pcx arrays to handle pcx files
RLECount	db  ?		;which are packed over lines
Position	dw  ?

MaxHeader	= sizeof.GifHeaderType	;Find maximum header size
if MaxHeader	< sizeof.BmpHeaderType
   MaxHeader	= sizeof.BmpHeaderType
end if
if MaxHeader	< sizeof.PcxHeaderType
   MaxHeader	= sizeof.PcxHeaderType
end if
if MaxHeader	< sizeof.TgaHeaderType
   MaxHeader	= sizeof.TgaHeaderType
end if
if MaxHeader	< sizeof.VocHeaderType
   MaxHeader	= sizeof.VocHeaderType
end if
if MaxHeader	< sizeof.WavHeaderType
   MaxHeader	= sizeof.WavHeaderType
end if

; use same Header buffer for all file types
virtual
GifHeader	GifHeaderType
end virtual
virtual
BmpHeader	BmpHeaderType
end virtual
virtual
PcxHeader	PcxHeaderType
end virtual
virtual
TgaHeader	TgaHeaderType
end virtual
virtual
VocHeader	VocHeaderType
end virtual
virtual
WavHeader	WavHeaderType
end virtual
Header		db  MaxHeader	dup(?)	;reseve maximum header size

InputBuffer	db  BUFFERSIZE	dup(?)
OutputBuffer	db  BUFFERSIZE	dup(?)
EndBuffer	= $		;GIF can't use this buffer !!!

; Gif variables/buffers outside other data
; Input Gif variables
GifBufPos	dw  ?
GifBufCount	dw  ?		;Bytes in ReBuffer
IClearCode	dw  ?		;Gif Clear code
IEofCode	dw  ?		;Gif EndOfFile code
IOldCode	dw  ?		;Decompressor variable
IFreeCode	dw  ?		;Decompressor, next free slot in hash table
IBitsLeft	dw  ?
ICodeBitsW	= $;label word
ICodeSize	db  ?		;Code size, read from Gif header
IBitCount	db  ?
InitCodeSize	db  ?		;Starting code size, used during Clear
IOldToken	db  ?		;Decompressor variable
; Output Gif variables
OLastChar	db  ?
OIsFirst	db  ?
OClearCode	dw  ?		;1 Shl InitCodeSize ;Clear code
OEofCode	dw  ?		;ClearCode + 1      ;End Of File marker
OFreeCode	dw  ?		;ClearCode + 2      ;First free code
OBitOffset	dw  ?
OPrefix 	dw  ?
OMaxCode	dw  ?
OCodeSize	dw  ?

gifi		GifImageBlock
gCon		GifControlBlock
ReBuffer	db  256 	dup(?)	;GifBuffer for read
WrBuffer	db  258 	dup(?)	;GifBuffer for write (+2 extra bytes)
; The hash table used by the decompressor
Prefix		dw  HASHSIZE	dup(?)
Suffix		db  HASHSIZE	dup(?)
; The hash table used by the compressor
hFirst		dw  HASHSIZE	dup(?)	;First entry with this value
hNext		dw  HASHSIZE	dup(?)	;Next entry along chain
hChar		db  HASHSIZE	dup(?)	;Suffix char

GifStack	= $
